package checkers.javari;

import java.lang.annotation.Annotation;

import javax.annotation.processing.ProcessingEnvironment;
import javax.lang.model.element.*;
import javax.lang.model.type.*;
import javax.lang.model.util.ElementFilter;
import javax.lang.model.util.Elements;

import java.util.*;

import checkers.quals.*;

import checkers.types.*;

import com.sun.source.tree.*;
import com.sun.tools.javac.tree.*;

/**
 * Adds support for the javari mutability type annotations to {@link
 * AnnotatedTypeFactory}. This means that the {@link
 * JavariAnnotatedTypeFactory#getClass} method will regard some
 * inputs as carrying the appropriate mutability annotation even if
 * they do not carry it explicitly. These include:
 *
 * <ul>
 *  <li>primitive types
 *  <li>inherited mutability annotations
 * </ul>
 */
public class JavariAnnotatedTypeFactory extends AnnotatedTypeFactory {

    /**
     * Creates an annotated type factory that treats certain inputs as having
     * javari mutability annotations even if they do not explicitly carry them.
     *
     * @param env the {@link ProcessingEnvironment} to use for tree and type
     *        utilities
     * @param root the root of the syntax tree (used for TreePaths)
     */
    public JavariAnnotatedTypeFactory(ProcessingEnvironment env, CompilationUnitTree root) {
        super(env, root);
    }

    /**
     * Returns the method overriden by the method passed as a parameter,
     * or null if there is none.
     *
     * @param methodElt an ExecutableElement for a method
     * @return an ExecutableElement representing the method directly
     * overriden by the argument, or null if none.
     */
    public ExecutableElement getOverridenMethod(ExecutableElement methodElt){

        TypeElement classElt = (TypeElement) methodElt.getEnclosingElement();
        TypeMirror classType = classElt.asType();
        List<? extends TypeMirror> supers
            = env.getTypeUtils().directSupertypes(classType);

        for(TypeMirror t : supers) {
            TypeElement superElt
                = (TypeElement) ((DeclaredType)t).asElement();
            List<ExecutableElement> methods
                = ElementFilter.methodsIn(superElt.getEnclosedElements());
            Elements eltUtils = env.getElementUtils();
            for (ExecutableElement e : methods) {
                if (eltUtils.overrides(methodElt, e, classElt))
                    return e;
            }
        }

        return null;
    }


    /**
     * Overshadows getClass in AnnotatedTypeFactory to provide:
     *
     * <ul>
     *  <li> special treatment for MemberSelectTree trees
     *  <li> mutability inheritance from enclosing element
     *  <li> resolve mutability of {@code @RoMaybe} method invocations
     * </ul>
     *
     * @param tree
     *            a tree node
     * @throws IllegalArgumentException
     *             if {@code tree} is null
     * @return the {@link AnnotatedClassType} corresponding to
     *         the given tree node and the annotations written on it
     */
    public AnnotatedClassType getClass(Tree tree) {

        if (tree instanceof MemberSelectTree) {
            MemberSelectTree mst = (MemberSelectTree)tree;
            AnnotatedClassType receiverType =
                super.getClass(mst);

            AnnotatedClassType outerType =
                getClass(mst.getExpression());

            Element outerElt = outerType.getElement();
            Element innerElt = null;
            {
                Name innerName = mst.getIdentifier();
                List<? extends Element> innerElts =
                    ElementFilter.fieldsIn(outerElt.getEnclosedElements());
                for (Element i : innerElts) {
                    if (i.getSimpleName().equals(innerName)) {
                        innerElt = i;
                        break;
                    }
                }
            }

            AnnotatedClassType innerType = getClass(innerElt);
            innerType.setElement(receiverType.getElement());
            innerType.setTree(tree);

            // if the inner type is specified, return it
            if (innerType.hasAnnotationAt(ReadOnly.class,
                                          AnnotationLocation.RAW)
                || innerType.hasAnnotationAt(Mutable.class,
                                             AnnotationLocation.RAW))
                return innerType;

            if (outerType.hasAnnotationAt(ReadOnly.class,
                                          AnnotationLocation.RAW))
                innerType.include(ReadOnly.class);
            else
                innerType.include(Mutable.class);

            return innerType;
        }


        AnnotatedClassType type = super.getClass(tree);

        if (tree instanceof IdentifierTree) {
            if (type.hasAnnotationAt(RoMaybe.class, AnnotationLocation.RAW)) {
                type.include(ReadOnly.class);
                return type;
            }
        } else if (tree instanceof LiteralTree) {
            type.include(Mutable.class);
            return type;
        } else if (tree instanceof BinaryTree) {
            type.include(Mutable.class);
            return type;

        } else if (tree instanceof MethodInvocationTree){

            MethodInvocationTree mit = (MethodInvocationTree) tree;
            if (type.hasAnnotationAt(RoMaybe.class,
                                     AnnotationLocation.RAW)){
                type = resolveRoMaybe(mit);
                return type;
            }

            if (!(type.hasAnnotationAt(ReadOnly.class,
                                       AnnotationLocation.RAW)
                  || type.hasAnnotationAt(Mutable.class,
                                          AnnotationLocation.RAW))) {
                if (mit.getMethodSelect() instanceof MemberSelectTree) {
                    MemberSelectTree mst =
                        (MemberSelectTree) (mit.getMethodSelect());
                    AnnotatedClassType referenceType =
                        getClass(mst.getExpression());

                    if (referenceType.hasAnnotationAt(ReadOnly.class,
                                                     AnnotationLocation.RAW))
                        type.include(ReadOnly.class);
                    else if (referenceType
                             .hasAnnotationAt(Mutable.class,
                                              AnnotationLocation.RAW))
                        type.include(Mutable.class);

                    return type;
                }
            }
        }

        if (!(type.hasAnnotationAt(ReadOnly.class, AnnotationLocation.RAW)
              || type.hasAnnotationAt(Mutable.class, AnnotationLocation.RAW))){
            Element elt = type.getElement();
            if (elt == null) {
                type.include(Mutable.class);
                return type;
            }
            Element enclosing = elt.getEnclosingElement();

            if (enclosing == null) {
                type.include(Mutable.class);
                return type;
            }

            if (getClass(enclosing).hasAnnotationAt(ReadOnly.class,
                                                    AnnotationLocation.RAW)) {
                type.include(ReadOnly.class);
                return type;
            }
            type.include(Mutable.class);
            return type;
        }

        return type;
    }

    // overriding so that every RoMaybe parameter is also ReadOnly
    protected List<AnnotatedClassType> paramTypes(ExecutableElement method) {

        List<AnnotatedClassType> paramTypes = super.paramTypes(method);
        for (AnnotatedClassType pType : paramTypes) {
            if (pType.hasAnnotationAt(RoMaybe.class, AnnotationLocation.RAW))
                pType.include(ReadOnly.class);
        }
        return paramTypes;
    }


    // determines whether a RoMaybe method is ReadOnly or not
    private AnnotatedClassType resolveRoMaybe(MethodInvocationTree tree) {
        AnnotatedMethodType method = getMethod(tree);
        ExecutableElement methodElt = method.getElement();
        RoMaybe roMaybe
            = (RoMaybe) methodElt.getAnnotation(RoMaybe.class);

        Set<String> returnKeys = new HashSet<String>();
        for (String s : roMaybe.value()) returnKeys.add(s);

        List<AnnotatedClassType> parameterTypes =
            method.getAnnotatedParameterTypes();

        Set<String>[] parameterKeys = new Set[parameterTypes.size()];

        for (int i = 0; i < parameterTypes.size(); i++) {
            Annotation a = parameterTypes.get(i).getElement()
                .getAnnotation(RoMaybe.class);
            if (a != null) {
                parameterKeys[i] = new HashSet<String>();
                for (String s : ((RoMaybe) a).value())
                    parameterKeys[i].add(s);
            }
        }

        boolean allRO = true;
        Set<String> collectedKeys = new HashSet<String>();

        List<? extends ExpressionTree> argumentTrees = tree.getArguments();

        for(int i = 0; i < parameterTypes.size(); i++) {
            if (parameterTypes.get(i)
                .hasAnnotationAt(RoMaybe.class,
                                 AnnotationLocation.RAW)){
                AnnotatedClassType argType
                    = getClass(argumentTrees.get(i));
                if (argType.hasAnnotationAt(ReadOnly.class,
                                             AnnotationLocation.RAW)){
                    collectedKeys.addAll(parameterKeys[i]);
                } else {
                    allRO = false;
                }
            }
        }


        AnnotatedClassType type = method.getAnnotatedReturnType();
        type.exclude(RoMaybe.class);

        if (returnKeys.isEmpty()) {
            if (allRO) {
                type.include(ReadOnly.class);
                type.exclude(Mutable.class);
                return type;
            }
            type.include(Mutable.class);
            type.exclude(ReadOnly.class);
            return type;
        } else {
            if (collectedKeys.containsAll(returnKeys)) {
                type.include(ReadOnly.class);
                type.exclude(Mutable.class);
                return type;
            }
            type.include(Mutable.class);
            type.exclude(ReadOnly.class);
            return type;
        }
    }
}
