package checkers.javari;

import java.io.*;
import java.lang.annotation.Annotation;
import java.util.*;
import javax.annotation.processing.SupportedAnnotationTypes;
import javax.annotation.processing.SupportedSourceVersion;
import javax.annotation.processing.ProcessingEnvironment;

import checkers.quals.ReadOnly;
import checkers.quals.Mutable;
import checkers.quals.RoMaybe;
import checkers.quals.QReadOnly;

import checkers.source.SourceVisitor;
import checkers.source.SourceChecker;
import checkers.types.*;

import com.sun.source.tree.CompilationUnitTree;
import javax.annotation.processing.SupportedAnnotationTypes;
import javax.annotation.processing.SupportedSourceVersion;
import javax.lang.model.SourceVersion;
import javax.lang.model.element.*;
import javax.lang.model.util.Elements;

/**
 * An annotation processor that checks a program's use of the Javari
 * mutability type annotations ({@code @ReadOnly} and {@code @Mutable}).
 */
@SupportedSourceVersion(SourceVersion.RELEASE_7)
@SupportedAnnotationTypes({"checkers.quals.ReadOnly",
                               "checkers.quals.Mutable",
                               "checkers.quals.QReadOnly",
                               "checkers.quals.RoMaybe",
                               "checkers.quals.Assignable"})
public class JavariChecker extends SourceChecker {

    @Override
    protected Properties getMessages(String fileName) throws IOException {

        // Use a message file if one exists.
        if (new File(fileName).exists()){
            String path = System.getProperty("messages", ".");
            Properties p = new Properties();
            p.load(new FileInputStream(path + File.separator + fileName));
            return p;
        }

        // Sets message defaults.
        Properties msgDefaults = new Properties();
        msgDefaults.setProperty("assignment.invalid", "cannot assign a ReadOnly expression to a Mutable variable");
        msgDefaults.setProperty("argument.invalid", "cannot pass ReadOnly argument as Mutable parameter");
        msgDefaults.setProperty("receiver.invalid", "cannot declare a method with Mutable receiver inside a ReadOnly type");
        msgDefaults.setProperty("return.invalid", "cannot return a ReadOnly value as Mutable");
        msgDefaults.setProperty("override.param.invalid", "cannot override Mutable parameter as ReadOnly");
        msgDefaults.setProperty("override.return.invalid", "cannot override Mutable return value as ReadOnly");
        msgDefaults.setProperty("override.throws.invalid", "cannot override Mutable throw type as ReadOnly");
        msgDefaults.setProperty("override.receiver.invalid", "cannot override Mutable receiver type as ReadOnly");
        msgDefaults.setProperty("mutable.cast", "cast increases mutability access");
        msgDefaults.setProperty("primitive.ro", "cannot declare primitives as ReadOnly");
        msgDefaults.setProperty("ro.field", "a field of a ReadOnly object is not assignable");
        msgDefaults.setProperty("ro.reference", "cannot invoke a Mutable method from a ReadOnly reference");
        msgDefaults.setProperty("ro.and.mutable", "a type cannot be declared as both ReadOnly and Mutable");

        return msgDefaults;
    }

    @Override
    public SourceVisitor getSourceVisitor(CompilationUnitTree root) {
        return new JavariVisitor(this, root);
    }

    /**
     * Overrides getFactory to produce a {@link JavariAnnotatedTypeFactory}.
     *
     * @param env
     *            the {@link ProcessingEnvironment} instance to use
     * @param root
     *            the root of the syntax tree that this factory produces
     *            annotated types for
     * @throws IllegalArgumentException
     *             if either argument is {@code null}.
     */
    @Override
    public AnnotatedTypeFactory getFactory(ProcessingEnvironment env, CompilationUnitTree root) {
        return new JavariAnnotatedTypeFactory(env, root);
    }


    /**
     * Determines whether the first type is a subtype of the second type
     * with respect to the annotations on raw types.  It ignores the actual
     * type and only checks annotations, and is only concerned with annotations
     * on raw types (it ignores all generic and array annotations).
     *
     * @param child the type that may or may not be a subtype
     * @param parent the parent type
     * @return true if {@code child} is a subtype of {@code parent} with respect
     *         to its type qualifier annotations, false otherwise
     */
    public boolean isSubtype(AnnotatedClassType child, AnnotatedClassType parent) {

        return isSubtypeAt(child, parent, AnnotationLocation.RAW);
    }


    private boolean isSubtypeAt(AnnotatedClassType child,
                                AnnotatedClassType parent,
                                AnnotationLocation location) {

        // only readonly from readonly
        if (child.hasAnnotationAt(ReadOnly.class, location)
            && !parent.hasAnnotationAt(ReadOnly.class, location))
            return false;

        // ? readonly can be assigned only to readonly or itself
        if (child.hasAnnotationAt(QReadOnly.class, location) &&
            !(parent.hasAnnotationAt(ReadOnly.class, location)
              || parent.hasAnnotationAt(QReadOnly.class, location)))
            return false;

        return true;
    }

    /**
     * Determines whether assignment is permissible from one type to another
     * given the rules of the type qualifier annotation being checked. It
     * accounts for generic and array annotations as well as annotations on the
     * raw type.
     *
     * @param from the right-hand type
     * @param to the left-hand type
     * @return true if {@code from} can be assigned to {@code to} with respect
     *         to the type qualifier annotation being checked
     */
    public boolean isAssignable(AnnotatedClassType from, AnnotatedClassType to) {

        // FIXME: this needs to check checkers.skipClasses!
        if (from.getElement() != null) {
            Elements elts = env.getElementUtils();
            Element fromElt = from.getElement();
            if (elts.getPackageOf(fromElt).getQualifiedName().toString().startsWith("java"))
                return true;
        }

        if (!isSubtype(from, to))
            return false;

        for (AnnotationLocation location : to.getAnnotatedLocations()) {
            if (!isSubtypeAt(from, to, location))
                return false;
        }

        for (AnnotationLocation location : from.getAnnotatedLocations()) {
            if (!isSubtypeAt(from, to, location))
                return false;
        }

        return true;
    }


    /**
     * Checks whether an AnnotatedClassType refers to a primitive type
     * marked as {@code @ReadOnly} (which is illegal).
     *
     * @param type the (@link AnnotatedClassType} to check
     * @return true if the type refers to a primitive and marked as
     * {@code @ReadOnly}.
     */
    public boolean isPrimitiveReadOnly(AnnotatedClassType type) {
        Element typeElt = type.getElement();
        if (typeElt instanceof VariableElement) {
            if (typeElt.asType().getKind().isPrimitive())
                return type.hasAnnotationAt(ReadOnly.class,
                                            AnnotationLocation.RAW);
        }
        return false;
    }

    /**
     * Checks whether an element refers to a primitive type.
     *
     * @param elt an Element to be verified
     * @return true if elt is instanceof (@link VariableElement} and
     * is primitive, false otherwise
     */
    public boolean isPrimitive(Element elt) {
        if (!(elt instanceof VariableElement)) return false;
        return (elt.asType().getKind().isPrimitive());
    }

}
