package checkers.javari;

import java.util.*;

import javax.lang.model.element.*;
import javax.lang.model.util.ElementFilter;

import java.lang.annotation.Annotation;

import checkers.source.*;
import checkers.subtype.*;
import checkers.types.*;

import checkers.javari.VisitorState;
import checkers.javari.VisitorState.State;

import checkers.quals.Mutable;
import checkers.quals.ReadOnly;
import checkers.quals.RoMaybe;
import checkers.quals.Assignable;
import checkers.quals.QReadOnly;

import com.sun.source.tree.*;
import com.sun.source.util.TreePath;

/**
 * A type-checking visitor for the Javari mutability annotations
 * ({@code @ReadOnly}, {@code @Mutable} and {@code @Assignable}) that
 * extends SourceVisitor.
 *
 * @see SourceVisitor
 */
public class JavariVisitor extends SourceVisitor<Void, VisitorState> {

    /** Whether or not system classes should not be checked against. */
    protected final boolean SKIP_SYSTEM_CLASSES =
            Boolean.valueOf(System.getProperty("checkers.skipSystemClasses", "true"));

    protected JavariChecker checker;

    protected TreePath rootPath;

    protected final Map<Element, Tree> treeMap = new HashMap<Element, Tree>();
    protected final Map<Tree, ClassTree> classMap =
        new HashMap<Tree, ClassTree>();
    protected final Map<Tree, MethodTree> methodMap =
        new HashMap<Tree, MethodTree>();


    /**
     * Creates a new visitor for type-checking the Javari mutability
     * annotations.
     *
     * @param checker the {@link JavariChecker} to use
     * @param root the root of the input program's AST to check
     */
    public JavariVisitor(JavariChecker checker, CompilationUnitTree root) {
        super(checker, root);
        this.checker = checker;
        rootPath = new TreePath(root);
        root.accept(new TreeMapperVisitor(treeMap, classMap, methodMap), null);
    }


    /**
     * Modifies p to the approriate values, according to super
     * classes, interfaces, and outer classes.
     *
     * @see com.sun.source.util.TreeScanner#visitClass(com.sun.source.tree.ClassTree,java.lang.Object)
     */
    @Override
    public Void visitClass(ClassTree node, VisitorState p) {
        boolean readonly = false, mutable = false;

        if (p == null){
            p = new VisitorState(null);
            mutable = true;
        }

        AnnotatedClassType type = factory.getClass(node);

        Tree ext = node.getExtendsClause();
        if (ext != null){
            AnnotatedClassType extendsType = factory.getClass(ext);
            if (extendsType.hasAnnotationAt(ReadOnly.class,
                                           AnnotationLocation.RAW))
                readonly = true;
            if (extendsType.hasAnnotationAt(Mutable.class,
                                            AnnotationLocation.RAW))
                mutable = true;
        }

        List<? extends Tree> imps = node.getImplementsClause();
        for(Tree it : imps) {
            AnnotatedClassType iType = factory.getClass(it);
            if (iType.hasAnnotationAt(ReadOnly.class,
                                      AnnotationLocation.RAW))
                readonly = true;
            if (iType.hasAnnotationAt(Mutable.class,
                                      AnnotationLocation.RAW))
                mutable = true;
        }

        if (type.hasAnnotationAt(ReadOnly.class, AnnotationLocation.RAW))
            readonly = true;
        else if (type.hasAnnotationAt(Mutable.class, AnnotationLocation.RAW))
            mutable = true;

        if (readonly && mutable) {
            checker.report(Result.failure("ro.and.mutable"), node);
            p.setClassState(State.MUTABLE);
        } else if (readonly)
            p.setClassState(State.READONLY);
        else
            p.setClassState(State.MUTABLE);

        return super.visitClass(node, p);
    }

    /**
     * Modifies the parameter p to its appropriate value. Checks:
     *
     * <ul>
     *  <li> the return type cannot be primitive and marked with ReadOnly
     *  <li> whether the method overrides a method illegally
     *  <li> whether the receiver type is invalid (Mutable inside ReadOnly)
     * </ul>
     *
     * @see com.sun.source.util.TreeScanner#visitMethod(com.sun.source.tree.MethodTree,java.lang.Object)
     */
    @Override
    public Void visitMethod(MethodTree node, VisitorState p) {

        AnnotatedMethodType method = factory.getMethod(node);

        ExecutableElement methodElt = method.getElement();

        AnnotatedClassType receiverType = method.getAnnotatedReceiverType();
        if (receiverType.hasAnnotationAt(ReadOnly.class,
                                         AnnotationLocation.RAW))
            p.state = State.READONLY;
        else if (receiverType.hasAnnotationAt(Mutable.class,
                                              AnnotationLocation.RAW))
            p.state = State.MUTABLE;
        else if (p.classState == State.MUTABLE)
            p.state = State.THIS_MUTABLE;
        else
            p.state = State.READONLY;

        // check if receiver is invalid
        if (receiverType.hasAnnotationAt(Mutable.class,
                                         AnnotationLocation.RAW)
            && p.classState == State.READONLY)
            checker.report(Result.failure("receiver.invalid"), node);

        // check if overriding is legal
        JavariAnnotatedTypeFactory jfactory
            = (JavariAnnotatedTypeFactory) factory;
        ExecutableElement oMethodElt = jfactory.getOverridenMethod(methodElt);

        if (oMethodElt != null) {

            AnnotatedMethodType oMethod = factory.getMethod(oMethodElt);

            // check parameters
            {
                List<AnnotatedClassType> mParamTypes
                    = method.getAnnotatedParameterTypes(),
                    omParamTypes = oMethod.getAnnotatedParameterTypes();
                for (int i = 0; i < mParamTypes.size(); i++) {
                    if (!checker.isAssignable(mParamTypes.get(i),
                                              omParamTypes.get(i)))
                        checker.report(Result
                                       .failure("override.param.invalid"),
                                       mParamTypes.get(i).getElement());
                }
            }
            // check throw types
            {
                List<AnnotatedClassType> mThrowsTypes
                    = method.getAnnotatedThrowsTypes(),
                    omThrowsTypes = oMethod.getAnnotatedThrowsTypes();
                for (int i = 0; i < mThrowsTypes.size(); i++) {
                    if (!checker.isAssignable(mThrowsTypes.get(i),
                                              omThrowsTypes.get(i)))
                        checker.report(Result
                                       .failure("override.throws.invalid"),
                                       mThrowsTypes.get(i).getElement());
                }
            }
            // check receiver
            {
                AnnotatedClassType mReceiverType
                    = method.getAnnotatedReceiverType(),
                    omReceiverType = oMethod.getAnnotatedReceiverType();
                if (!checker.isAssignable(mReceiverType, omReceiverType))
                    checker.report(Result
                                   .failure("override.receiver.invalid"),
                                   mReceiverType.getElement());
            }
            // check return
            {
                AnnotatedClassType mReturnType
                    = method.getAnnotatedReturnType(),
                    omReturnType = oMethod.getAnnotatedReturnType();
                if (!checker.isAssignable(mReturnType, omReturnType))
                    checker.report(Result
                                   .failure("override.return.invalid"),
                                   mReturnType.getElement());
            }

        }

        if (node.getReturnType() instanceof PrimitiveTypeTree) {
            AnnotatedClassType returnType = method.getAnnotatedReturnType();
            if (returnType.hasAnnotationAt(ReadOnly.class,
                                           AnnotationLocation.RAW)) {
                checker.report(Result.failure("primitive.ro"),
                               node.getReturnType());
            }
        }

        return super.visitMethod(node, p);

    }

    /**
     * Checks whether the assignment is possible:
     *
     * <ul>
     *  <li> ReadOnly references may not be assigned to Mutable variables
     *  <li> ReadOnly fields are not assignable, unless marked with
     *  {@code @Assignable}
     *  <li> local fields are final in a READONLY context
     * </ul>
     *
     * @see com.sun.source.util.TreeScanner#visitAssignment(com.sun.source.tree.AssignmentTree,java.lang.Object)
     */
    @Override
    public Void visitAssignment(AssignmentTree node, VisitorState p) {

        commonAssignment(node.getExpression(), node.getVariable(), p);
        return super.visitAssignment(node, p);
    }


    /**
     * Same implementation of  visitAssignment, but for {@link CompoundAssignmentTree}.
     *
     * @see com.sun.source.util.TreeScanner#visitCompoundAssignment(com.sun.source.tree.CompoundAssignmentTree,java.lang.Object)
     */
    @Override
    public Void visitCompoundAssignment(CompoundAssignmentTree node,
                                        VisitorState p) {
        commonAssignment(node.getExpression(), node.getVariable(), p);
        return super.visitCompoundAssignment(node, p);
    }


    // code for assignments
    private void commonAssignment(ExpressionTree expressionTree,
                                  ExpressionTree variableTree,
                                  VisitorState p){

        AnnotatedClassType expression = factory.getClass(expressionTree);
        AnnotatedClassType variable = factory.getClass(variableTree);

        // must check if can be used as readonly
        if (p.state == State.READONLY) {
            if (!variable.hasAnnotationAt(Assignable.class,
                                          AnnotationLocation.RAW)
                && ElementFilter
                .fieldsIn(InternalUtils.symbol(classMap.get(expressionTree))
                          .getEnclosedElements())
                .contains(variable.getElement())) { // if thing is local field
                checker.report(Result.failure("ro.field"),
                               variableTree);
            }
        }

        if (variableTree instanceof MemberSelectTree) {
            if ((variable.hasAnnotationAt(ReadOnly.class,
                                          AnnotationLocation.RAW)
                 || variable.hasAnnotationAt(QReadOnly.class,
                                             AnnotationLocation.RAW))
                && !variable.hasAnnotationAt(Assignable.class,
                                             AnnotationLocation.RAW))
                checker.report(Result.failure("ro.field"),
                               variableTree);
        } else {
            // not a field, so can set to a new object
            if (expressionTree instanceof NewClassTree)
                return;
        }

        if (!checker.isAssignable(expression, variable))
            checker.report(Result.failure("assignment.invalid"),
                           expressionTree);
    }


    /**
     * Checks for prefix and posfix increments and decrements on final fields.
     *
     * @see com.sun.source.util.TreeScanner#visitUnary(com.sun.source.tree.UnaryTree,java.lang.Object)
     */
    public Void visitUnary(UnaryTree node, VisitorState p){
        Tree.Kind kind = node.getKind();
        if (p.state != State.MUTABLE) {
            if (kind == Tree.Kind.POSTFIX_DECREMENT
                || kind == Tree.Kind.POSTFIX_INCREMENT
                || kind == Tree.Kind.PREFIX_DECREMENT
                || kind == Tree.Kind.PREFIX_INCREMENT) {

                AnnotatedClassType variable
                    = factory.getClass(node.getExpression());
                if (!variable.hasAnnotationAt(Assignable.class,
                                              AnnotationLocation.RAW)
                    && ElementFilter
                    .fieldsIn(InternalUtils.symbol(classMap.get(node))
                              .getEnclosedElements())
                    .contains(variable.getElement())) { // if local field

                    checker.report(Result.failure("ro.field"),
                                   node.getExpression());
                }
            }
        }
        return super.visitUnary(node, p);
    }

    /**
     * Emits a warning if a cast is made from readonly to mutable.
     *
     * @see com.sun.source.util.TreeScanner#visitTypeCast(com.sun.source.Tree.TypeCastTree,java.lang.Object)
     */
    public Void visitTypeCast(TypeCastTree node, VisitorState p) {
        AnnotatedClassType exprType = factory.getClass(node.getExpression());
        AnnotatedClassType castType = factory.getClass(node.getType());

        if (!checker.isAssignable(exprType, castType))
            checker.report(Result.warning("mutable.cast"), node.getType());

        return super.visitTypeCast(node, p);
    }


    /**
     * Checks if the variable is not a ReadOnly primitive, and if it
     * is initialized, whether the initial value is assignable.
     *
     * @see com.sun.source.util.TreeScanner#visitVariable(com.sun.source.tree.VariableTree,java.lang.Object)
     */
    @Override
    public Void visitVariable(VariableTree node, VisitorState p){
        AnnotatedClassType variable = factory.getClass(node);

        if (checker.isPrimitiveReadOnly(variable))
            checker.report(Result.failure("primitive.ro"), node);

        ExpressionTree initializer = node.getInitializer();

        if (initializer == null || (initializer instanceof NewClassTree))
            return super.visitVariable(node, p);

        AnnotatedClassType init = factory.getClass(initializer);

        if (!checker.isAssignable(init, variable))
            checker.report(Result.failure("assignment.invalid"), variable
                           .getElement());

        return super.visitVariable(node, p);
    }


    /**
     * Checks if a method invocation satisfies the following conditions:
     *
     *<ul>
     * <li> All arguments are assignable to its parameter types.
     * <li> A mutable method cannot be invoked from inside a method
     * that is not mutable.
     * <li> A mutable method cannot be invoked from a readonly reference.
     *</ul>
     *
     * Also responsible for scanning the invoked method to check if it
     * is mutable despite being annotated as thismutable.
     *
     * @see com.sun.source.util.TreeScanner#visitMethodInvocation(com.sun.source.tree.MethodInvocationTree,java.lang.Object)
     */
    @Override
    public Void visitMethodInvocation(MethodInvocationTree node,
                                      VisitorState p) {

        AnnotatedMethodType method = factory.getMethod(node);
        // need to ask factory to resolve RoMaybes
        AnnotatedClassType returnType = factory.getClass(node);

        Element methodElt = method.getElement();

        if (node.getMethodSelect() instanceof MemberSelectTree) {
            MemberSelectTree mst = (MemberSelectTree) (node.getMethodSelect());

            AnnotatedClassType expressionType =
                factory.getClass(mst.getExpression());
            AnnotatedClassType receiverType
                = method.getAnnotatedReceiverType();

            if (expressionType.hasAnnotationAt(ReadOnly.class,
                                               AnnotationLocation.RAW)) {
                if (receiverType.hasAnnotationAt(Mutable.class,
                                                 AnnotationLocation.RAW))
                    checker.report(Result.failure("ro.reference"), node);
            }
        }


        if (p.state == State.READONLY) {
            AnnotatedClassType receiver = method.getAnnotatedReceiverType();
            if (receiver.hasAnnotationAt(Mutable.class,
                                         AnnotationLocation.RAW)) {
                checker.report(Result.failure("ro.reference"), node);
            }
        }


        List<AnnotatedClassType> parameters = method
                .getAnnotatedParameterTypes();

        List<AnnotatedClassType> arguments
            = new LinkedList<AnnotatedClassType>();

        for (ExpressionTree arg : node.getArguments())
            arguments.add(factory.getClass(arg));

        for (int i = 0; i < arguments.size(); i++) {
            if (!checker.isSubtype(arguments.get(i), parameters.get(i)))
                checker.report(Result.failure("argument.invalid"), node);
        }

        return super.visitMethodInvocation(node, p);
    }

    /**
     * Checks whether the return type of a returned variable is
     * assignable to its declared return type.
     *
     * @see com.sun.source.util.TreeScanner#visitReturn(com.sun.source.tree.ReturnTree,java.lang.Object)
     */
    // (The second parameter in the above @see expression is Void, not Object,
    // but the javadoc tool fails to recognize it with the correct type.)
    @Override
    public Void visitReturn(ReturnTree node, VisitorState p) {

        if (node.getExpression() == null)
            return super.visitReturn(node, p);

        AnnotatedClassType ret = factory.getClass(node.getExpression());
        Tree enclosingMethod = InternalUtils.enclosingMethod(TreePath.getPath(
                root, node));

        AnnotatedMethodType meth = factory.getMethod(enclosingMethod);

        if (!checker.isAssignable(ret, meth.getAnnotatedReturnType()))
            checker.report(Result.failure("return.invalid"), node);

        return super.visitReturn(node, p);
    }

}
