package checkers.nonnull;

import checkers.types.*;

import com.sun.source.tree.*;
import com.sun.source.util.*;

import java.util.*;

import javax.annotation.processing.*;
import javax.lang.model.element.*;


/**
 * Performs a flow-sensitive {@code @NonNull} analysis on a portion of an AST.
 * Specifically, it checks if conditionals to determine if there is an explicit
 * null check (e.g., "if (x != null)"), and it determines a scope for which the
 * checked variable is nonnull. For local variables, the scope ends where the
 * variable is first reassigned; for fields and method parameters, the scope
 * ends where the variable is first reassigned or where it is first passed as
 * an argument to a method. 
 */
class FlowVisitor extends TreePathScanner<Void, Void> {

    /** The root of the source tree (required for pathfinding). */
    private final CompilationUnitTree root;

    /** Used to determine source positions for variable scopes. */
    private final SourcePositions srcPos;

    /** A stack of scopes (for nested conditionals). */
    private final Deque<FlowScope> scopes = new ArrayDeque<FlowScope>();

    /** The set of completed scopes. */
    private final Set<FlowScope> results = new HashSet<FlowScope>();

    /**
     * Creates a visitor for performing flow-sensitive nonnull analysis on a
     * subtree of the given tree root. It uses the provided {@link
     * SourcePositions} to determine a range of positions for which an element
     * is nonnull.
     * 
     * @param root the root of the subtree that will be checked
     * @param srcPos {@link SourcePositions} for producing ranges
     */
    public FlowVisitor(CompilationUnitTree root, SourcePositions srcPos) {
        this.root = root;
        this.srcPos = srcPos;
    }
    
    /**
     * Retrieves the results of the completed analysis.
     *
     * @return an unmodifiable set of completed scopes
     */
    public Set<? extends FlowScope> getResults() {
        return Collections.unmodifiableSet(results);
    }
    
    private ExpressionTree getOperandNotNull(BinaryTree tree) {
        if (tree.getLeftOperand().getKind() == Tree.Kind.NULL_LITERAL)
            return tree.getRightOperand();
        else if (tree.getRightOperand().getKind() == Tree.Kind.NULL_LITERAL)
            return tree.getLeftOperand();
        else return null;
    }
    
    @Override
    public Void visitIf(IfTree node, Void p) {

        // If the condition of this if is not a binary tree, return.
        ExpressionTree expr = node.getCondition();
        if (!(expr instanceof BinaryTree))
            return super.visitIf(node, p);

        BinaryTree condition = (BinaryTree)expr;

        // scopeStatement is either the "then" condition or the "else"
        // condition, depending on whether it's an "if (x != null)" check or an
        // "if (x == null)" check, respectively.
        //
        // nnExpression is the variable that the check makes nonnull, e.g., "x"
        // in "if (x != null)". 
        
        StatementTree scopeStatement;
        ExpressionTree nnExpression;

        switch (condition.getKind()) {
            case NOT_EQUAL_TO:
                scopeStatement = node.getThenStatement();
                nnExpression = this.getOperandNotNull(condition);
                break;

            case EQUAL_TO:
                scopeStatement = node.getElseStatement();
                nnExpression = this.getOperandNotNull(condition);
                break;

            default:
                scopeStatement = null;
                nnExpression = null;
        }
            
        // Return if it wasn't an equal-to or not-equal-to check against the
        // null literal.
        if (nnExpression == null)
            return super.visitIf(node, p);

        // Push a new FlowScope for the checked variable that starts at the
        // beginning of scopeStatement.
        Element nnElement = InternalUtils.symbol(nnExpression);
        FlowScope scope = new FlowScope(nnElement,
                srcPos.getStartPosition(root, scopeStatement));
        scopes.push(scope);

        // Continue scanning the if statement's "then" and "else" blocks.
        super.visitIf(node, p);

        // Pop a scope from the stack -- it should be same as the one we
        // pushed.
        FlowScope done = scopes.pop();
        assert scope == done;

        // If the scope hasn't previously been completed (by reaching an
        // assignment (for all variables) or a method invocation (for
        // non-locals)), complete it with the end of scopeStatement (i.e.,
        // meaning that the checked variable is nonnull for the entire "then"
        // or "else" block.
        if (!scope.isComplete()) {
            // Workaround when blocks aren't used.
            if (scopeStatement instanceof ExpressionStatementTree)
                scope.complete(Long.MAX_VALUE);
            else scope.complete(srcPos.getEndPosition(root, scopeStatement));
        }

        // Add the completed scope to the result.
        results.add(scope);

        return null;
    }

    @Override
    public Void visitAssignment(AssignmentTree node, Void p) {

        FlowScope scope = scopes.peek();

        // If the most recent scope hasn't been completed and the checked
        // variable is being assigned to, complete the scope just before the
        // assignment.
        if (scope != null && !scope.isComplete()) {
            Element var = InternalUtils.symbol(node.getVariable());
            if (scope.getElement().equals(var))
                scope.complete(srcPos.getStartPosition(root, node.getVariable()) - 1);
        }

        return super.visitAssignment(node, p);
    }

    public Void visitMethodInvocation(MethodInvocationTree node, Void p) {

        FlowScope scope = scopes.peek();

        // If the most recent scope hasn't been completed and there's a method
        // invocation (and the scope element isn't a local variable), complete
        // the scope just before the method invocation.
        if (scope != null && !scope.isComplete() 
                && scope.getElement().getKind() != ElementKind.LOCAL_VARIABLE)
            scope.complete(srcPos.getStartPosition(root, node) - 1);
        
        return super.visitMethodInvocation(node, p);
    }
}
