package checkers.source;

import java.util.*;
import checkers.quals.*;

/**
 * Represents the outcome of a typechecking operation (success, warning, or
 * failure, plus a list of explanatory messages). {@link Result}s created
 * during typechecking can be reported using {@link SourceChecker#report}, which
 * ultimately delivers an error or warning message via the JSR 199 compiler
 * interface.
 *
 * @see SourceChecker#report
 */
public final class Result {

    private static enum Type {
        SUCCESS, FAILURE, WARNING;

        /**
         * @return whichever of the given types is most serious
         */
        public static final Type merge(Type a, Type b) {
            if (a == FAILURE || b == FAILURE)
                return FAILURE;
            else if (a == WARNING || b == WARNING)
                return WARNING;
            else
                return SUCCESS;
        }
    };

    /** The type of result (success, warning, failure). */
    private final Type type;

    /** The message keys for this result. */
    private final List<String> messageKeys;

    /** The success result. */
    public static final Result SUCCESS = new Result(Type.SUCCESS, null);

    /**
     * Creates a new failure result with the given message key.
     * 
     * @param messageKey
     *            the key representing the reason for failure
     * @return the failure result
     */
    public static Result failure(/*@NonNull*/ String messageKey) {
        return new Result(Type.FAILURE, Collections.singletonList(messageKey));
    }

    /**
     * Creates a new warning result with the given message.
     * 
     * @param messageKey
     *            the key for the warning message
     * @return the warning result
     */
    public static Result warning(/*@NonNull*/ String messageKey) {
        return new Result(Type.WARNING, Collections.singletonList(messageKey));
    }

    /**
     * Creates a non-success result.
     * 
     * @param type
     *            the type of result
     * @param messageKey
     *            the message key for this result
     */
    private Result(Type type, List<String> messageKeys) {
        this.type = type;
        this.messageKeys = new LinkedList<String>();
        if (messageKeys != null)
            this.messageKeys.addAll(messageKeys);
    }

    /**
     * Merges two results into one.
     * 
     * @param r
     *            the result to merge with this result
     * @return a result that is the success result if both this and {@code r}
     *         are success results, or a result that has the more significant
     *         type (failure > warning > success) and the message keys of both
     *         this result and {@code r}
     */
    public Result merge(Result r) {
        if (r == null)
            return this;

        if (r.isSuccess() && this.isSuccess())
            return SUCCESS;

        List<String> messages = new LinkedList<String>();
        messages.addAll(this.messageKeys);
        messages.addAll(r.messageKeys);
        return new Result(Type.merge(r.type, this.type), messages); 
                
    }

    /**
     * @return true if the result is success (not a failure or warning)
     */
    public boolean isSuccess() {
        return type == Type.SUCCESS;
    }

    /**
     * @return true if the result is a failure
     */
    public boolean isFailure() {
        return type == Type.FAILURE;
    }

    /**
     * @return true if the result is a warning
     */
    public boolean isWarning() {
        return type == Type.WARNING;
    }

    /**
     * @return the message keys associated with the result
     */
    public List<String> getMessageKeys() {
        return Collections.unmodifiableList(messageKeys);
    }

    @Override
    public String toString() {
        switch (type) {
        case FAILURE:
            return "FAILURE: " + messageKeys;
        case WARNING:
            return "WARNING: " + messageKeys;
        case SUCCESS:
        default:
            return "SUCCESS";
        }
    }
}
