package checkers.subtype;

import java.io.*;
import java.lang.annotation.Annotation;
import java.util.*;
import javax.annotation.processing.SupportedAnnotationTypes;
import javax.annotation.processing.SupportedSourceVersion;

import javax.lang.model.SourceVersion;
import javax.lang.model.element.*;
import javax.lang.model.util.Elements;

import checkers.source.*;
import checkers.types.*;

import com.sun.source.tree.CompilationUnitTree;

/**
 * An annotation processor for typechecking qualifiers for which the qualified
 * type is the subtype of the unqualified type.
 *
 * <p>
 *
 * This class is intended for subclassing. It provides two primary services:
 * <ul>
 *  <li>{@link SubtypeVisitor}, an implementation of {@link SourceVisitor} that
 * scans programs for assignability violations with respect to the subtype
 * qualifier (i.e., if the qualified type of @A is the subtype of the
 * qualified type, the visitor checks that data with the unqualified type is
 * never assigned to variables with the qualified type)</li>
 *  <li>the {@link SubtypeChecker#isSubtype} and {@link SubtypeChecker#isAssignable}
 * helper methods for {@link SubtypeVisitor} and its potential subclasses</li>
 * </ul>
 */
@SupportedSourceVersion(SourceVersion.RELEASE_7)
@SupportedAnnotationTypes({"*"})
@SupportedLintOptions({"cast", "cast:redundant"})
public abstract class SubtypeChecker extends SourceChecker {

    /** The annotation to check (in isSubtype and isAssignable). */
    protected final Class<? extends Annotation> annotation;

    /**
     * Creates an annotation processor for checking subtype qualifiers. The
     * processor will perform checking using the annotation class named by the
     * value of the "annotation" JVM property, which may be set using the
     * "-Dannotation=[value]" command-line switch.
     */
    @SuppressWarnings("unchecked")
    public SubtypeChecker() {
        String annotationName = System.getProperty("annotation");

        try {
            this.annotation = (Class<? extends Annotation>)Class.forName(annotationName);
        } catch (ClassNotFoundException e) {
            throw new IllegalArgumentException("invalid annotation: " + annotationName);
        }
    }

    /**
     * Creates an annotation processor for checking subtype qualifiers.
     *
     * @param annotation the class of the annotation that this annotation
     * processor will check
     */
    public SubtypeChecker(Class<? extends Annotation> annotation) {
        this.annotation = annotation;
    }

    @Override
    protected/* @NonNull */Properties getMessages(String fileName) throws IOException {
        // Read messages from a file (in the working directory, by default, or
        // else in a directory indicated by the system "messages" property.
        String path = System.getProperty("messages", ".");
        Properties p = new Properties();
        p.load(new FileInputStream(path + File.separator + fileName));
        return p;
    }

    /**
     * Determines whether the first type is a subtype of the second type
     * with respect to the annotations on raw types.  It ignores the actual
     * type and only checks annotations, and is only concerned with annotations
     * on raw types (it ignores all generic and array annotations).
     *
     * @param child the type that may or may not be a subtype
     * @param parent the parent type
     * @return true if {@code child} is a subtype of {@code parent} with respect
     *         to its type qualifier annotations, false otherwise
     */
    protected boolean isSubtype(AnnotatedClassType child, AnnotatedClassType parent) {
        if (!child.hasAnnotationAt(this.annotation, AnnotationLocation.RAW) &&
                parent.hasAnnotationAt(this.annotation, AnnotationLocation.RAW))
            return false;
        return true;
    }

    /**
     * Determines whether assignment is permissible from one type to another
     * given the rules of the type qualifier annotation being checked. It
     * accounts for generic and array annotations as well as annotations on the
     * raw type.
     *
     * @param from the right-hand type
     * @param to the left-hand type
     * @return true if {@code from} can be assigned to {@code to} with respect
     *         to the type qualifier annotation being checked
     */
    public boolean isAssignable(AnnotatedClassType from, AnnotatedClassType to) {

        if (from.getElement() != null) {
            Elements elts = env.getElementUtils();
            Element fromElt = from.getElement();

            String className = InternalUtils.getQualifiedName(fromElt);
            if (this.shouldSkip(className))
                return true;
        }

        if (!isSubtype(from, to))
            return false;

        boolean invalid = false;

        for (AnnotationLocation location : to.getTypeArgumentLocations()) {
            if (!from.hasAnnotationAt(this.annotation, location)
                    && !from.hasWildcardAt(location))
                return false;
        }

        for (AnnotationLocation location : from.getTypeArgumentLocations()) {
            if (!to.hasAnnotationAt(this.annotation, location)
                    && !to.hasWildcardAt(location))
                return false;
        }

        return true;
    }
}
