package checkers.types;

import java.util.*;

/**
 * Represents the location of an annotation within a type.  For example, consider
 *
 * <pre><code>@A HashMap&lt;@B String, @C List &lt;@D Object&gt;&gt;</code></pre>
 *
 * The <code>A</code> annotation is on the top-level type, the
 * <code>B</code> and <code>C</code> annotations are at the first level of
 * generic type arguments, and the <code>D</code> annotation is at the
 * second level of generic type arguments.
 * <p>
 *
 * AnnotationLocation is a wrapper for a list of integers; the <i>i</i>th
 * integer in the list holds the index (from zero) of a parameter at level
 * <i>i</i> (again, counting from zero) in the generic type.  Here are the
 * representations of the annotations given above:
 *
 * <table>
 * <tr><th> A <td>&nbsp;{@link AnnotationLocation#RAW} = []
 * <tr><th> B <td>&nbsp;[ 0 ]
 * <tr><th> C <td>&nbsp;[ 1 ]
 * <tr><th> D <td>&nbsp;[ 1, 0 ]
 * </table>
 *
 */
public final class AnnotationLocation {

    /** The location for an annotation on a raw type. */
    public static final AnnotationLocation RAW = new AnnotationLocation(
            new int[0]);

    /** The integer list specifying the location (see above). */
    private final List<Integer> location;

    /**
     * Returns an annotation location for the given array. This method will
     * return {@link AnnotationLocation#RAW} for arrays of length zero.
     *
     * @return the {@link AnnotationLocation} for the given array
     */
    public static AnnotationLocation fromArray(int[] location) {
        if (Arrays.equals(location, new int[0]))
            return RAW;
        else return new AnnotationLocation(location);
    }
    
    /**
     * Returns an annotation location for the given list. This method will
     * return {@link AnnotationLocation#RAW} for lists of size zero.
     *
     * @return the {@link AnnotationLocation} for the given list
     */
    public static AnnotationLocation fromList(List<Integer> location) {
        if (location.size() == 0)
            return RAW;
        else return new AnnotationLocation(location);               
    }
    
    /**
     * Creates a new instance of {@link AnnotationLocation}.
     *
     * @param location
     *            the integer list specifying the location (see above)
     */
    private AnnotationLocation(int[] location) {

        ArrayList<Integer> lst = new ArrayList<Integer>(location.length);
        for (int i : location)
            lst.add(i);

        this.location = Collections.unmodifiableList(lst);
    }

    /**
     * Creates a new instance of {@link AnnotationLocation}.
     *
     * @param location
     *            the integer list specifying the location (see above)
     */
    private AnnotationLocation(List<Integer> location) {
        this.location = Collections.unmodifiableList(location);
    }

    /**
     * Returns the location as a list of indices in the generic type.
     *
     * @return the integer list specifying the location (see above)
     */
    public List<Integer> asList() {
        return location;
    }

    @Override
    public boolean equals(Object o) {
        return (o instanceof AnnotationLocation && 
                ((AnnotationLocation) o).location.equals(location));
    }

    @Override
    public int hashCode() {
        return location.hashCode() + 37;
    }

    @Override
    public String toString() {
        return location.toString();
    }
}
