package checkers.types;

import java.lang.annotation.Annotation;
import java.util.*;

import javax.lang.model.element.*;

import com.sun.source.tree.Tree;
import com.sun.tools.javac.code.*;
import com.sun.tools.javac.code.Symbol.*;
import com.sun.tools.javac.comp.*;

/**
 * Represents an annotation along with its annotation target (its
 * {@link AnnotationTarget}), which cannot normally be obtained from an
 * {@link AnnotationMirror}. Some methods in this class depend on non-public
 * APIs.
 *
 * @see AnnotationTarget
 */
final class InternalAnnotation {

    /** The annotation's Target. */
    private final AnnotationTarget target;

    /** The annotation's location. */
    private final AnnotationLocation location;

    /** The annotation's class. */
    private final Class<? extends Annotation> annotation;

    /**
     * Creates a new {@link InternalAnnotation}.
     *
     * @param target
     *            the annotation's target
     * @param annotation
     *            the annotation's class
     */
    private InternalAnnotation(AnnotationTarget target,
            Class<? extends Annotation> annotation) {
        this.target = target;
        if (target == null)
            this.location = AnnotationLocation.RAW;
        else if (target.type.hasLocation())
            this.location = AnnotationLocation.fromArray(target.location);
        else
            this.location = AnnotationLocation.RAW;
        this.annotation = annotation;
    }

    /**
     * Determines whether or not the annotation is an extended annotation.
     *
     * @return true if the represented annotation is an extended annotation,
     *         false otherwise
     */
    public boolean isExtended() {
        return this.target != null && this.target.type != TargetType.UNKNOWN;
    }

    /**
     * @return the target for this extended annotation
     */
    public AnnotationTarget getTarget() {
        return this.target;
    }

    /**
     * Determines whether or not the annotation is on a raw type.
     *
     * @return true if the annotation is on a raw type, false if it's on a
     *         type argument (generic or array)
     */
    public boolean isRaw() {
        assert this.location == AnnotationLocation.RAW;
        return !this.target.type.hasLocation();
    }

    /**
     * @return the location of the annotation
     */
    public AnnotationLocation getLocation() {
        return this.location;
    }

    /**
     * @return the extended annotation's class
     */
    public Class<? extends Annotation> getAnnotationClass() {
        return this.annotation;
    }

    /**
     * Produces an {@link InternalAnnotation} from an {@link AnnotationMirror}.
     *
     * @param a
     *            the {@link AnnotationMirror} to use to construct the {@link
     *            InternalAnnotation}
     *
     * @return the {@link InternalAnnotation} corresponding to the given {@link
     *         AnnotationMirror}
     */
    @SuppressWarnings("unchecked")
    public static InternalAnnotation fromMirror(AnnotationMirror a) {
        Class<? extends Annotation> clazz;

        // Attempt to get the class of the annotation.
        try {
            clazz = (Class<Annotation>) Class.forName(a.getAnnotationType()
                    .toString());
        } catch (ClassNotFoundException e) {
            throw new IllegalArgumentException("invalid annotation class");
        }

        // Construct an InternalAnnotation from the attribute's target.
        if (a instanceof Attribute.Compound) {
            Attribute.Compound attr = (Attribute.Compound) a;

            if (attr.target != null) { // if it's extended
        	return new InternalAnnotation(attr.target, clazz);
            }
        }
        return new InternalAnnotation(null, clazz);
    }

    /**
     * Creates a list of the annotations from the given element; annotations
     * that were manually added are not included.
     *
     * @param element
     *            the {@link Element} to get annotations from
     *
     * @return the element's annotations as {@link InternalAnnotation}s
     */
    public static List<InternalAnnotation> fromElement(Element element) {
        List<InternalAnnotation> annotations = new LinkedList<InternalAnnotation>();

        if (element == null)
            return annotations;

        if (!(element instanceof Symbol))
            throw new RuntimeException("element not a symbol");

        Symbol s = (Symbol) element;

        assert s.attributes_field != null;

        for (AnnotationMirror a : s.attributes_field)
            annotations.add(InternalAnnotation.fromMirror(a));

        return annotations;
    }

    /**
     * Creates a list of the annotations in the
     * given {@link Element} that reference the given {@link Tree}. Useful for
     * annotations that are located in the body of the enclosing method.
     *
     * @param e
     *            the source of the annotations
     * @param tree
     *            the {@link Tree} that the annotations in {@code e} belong to
     *
     * @return the annotations on {@code e} that refer to {@code tree}
     */
    public static List<InternalAnnotation> fromElement(Element e,
            Tree tree) {

        List<InternalAnnotation> methodAnnos = fromElement(e);
        List<InternalAnnotation> thisAnnos = new LinkedList<InternalAnnotation>();

        for (InternalAnnotation a : methodAnnos) {
            AnnotationTarget target = a.getTarget();
            if (target == null || target.ref == null)
                continue;
            if (InternalUtils.refersTo(target.ref, tree))
                thisAnnos.add(a);
        }

        return thisAnnos;
    }

    /**
     * For each method parameter in the given element, gathers the annotations
     * on that parameter into a list and returns these lists.
     *
     * @param element
     *            the (symbol of the) method from which parameter annotations
     *            will be extracted
     * @return a list of the lists of the {@InternalAnnotation}s for each
     *         parameter in the method given by {@code element}
     */
    public static List<List<InternalAnnotation>> forMethodParameters(
            Element element) {

        List<List<InternalAnnotation>> annotations = new LinkedList<List<InternalAnnotation>>();

        if (element instanceof MethodSymbol)
            for (VarSymbol vs : ((MethodSymbol) element).params())
                annotations.add(fromElement(vs));

        return annotations;
    }

}
